# type: ignore
import cython
from cython.cimports.cpython import array


def levenshtein(s: str, t: str) -> int:
    # https://turnerj.com/blog/levenshtein-distance-part-2-gotta-go-fast
    previous_diagonal: cython.uint
    previous_column: cython.uint
    local_cost: cython.uint
    deletion_cost: cython.uint
    i: cython.uint
    j: cython.uint
    if s == t:
        return 0
    l_s: cython.uint = len(s)
    l_t: cython.uint = len(t)
    if l_s == 0:
        return l_t
    if l_t == 0:
        return l_s
    previous_row: cython.uint[:] = array.array("I", range(l_t + 1))
    for i in range(1, l_s + 1):
        previous_diagonal = i - 1
        previous_column = previous_diagonal + 1
        with cython.boundscheck(False), cython.wraparound(False):
            sourceChar = s[i - 1]
        for j in range(1, l_t + 1):
            local_cost = previous_diagonal
            deletion_cost = previous_row[j]
            if sourceChar != t[j - 1]:
                local_cost = min(local_cost, deletion_cost, previous_column) + 1
            previous_column = local_cost
            previous_row[j] = local_cost
            previous_diagonal = deletion_cost
    return previous_row[j]



from jellyfish import levenshtein_distance as levenshtein_jelly
from fxkk import (
    levenshtein_mat_py,
    levenshtein_vec_py,
    levenshtein_exp_py,
    levenshtein_cy,
    levenshtein_mat_rs,
    levenshtein_vec_rs,
    levenshtein_tweaked_rs,
)
from typing import Any, Dict, List

cases: List[Dict[str, Any]] = [
    {"a": "flaw", "b": "lawn", "expected": 2},
    {"a": "intention", "b": "execution", "expected": 5},
    {"a": "giraffe", "b": "gorilla", "expected": 5},
    {"a": "book", "b": "back", "expected": 2},
    {"a": "apple", "b": "apricot", "expected": 5},
    {"a": "hello", "b": "hallo", "expected": 1},
    {"a": "algorithm", "b": "altruistic", "expected": 6},
    {"a": "abcdefg", "b": "abcdxyz", "expected": 3},
    {"a": "mouse", "b": "mouses", "expected": 1},
    {"a": "sunday", "b": "saturday", "expected": 3},
    {"a": "дом", "b": "том", "expected": 1},
    {"a": "привет", "b": "пока", "expected": 5},
    {"a": "молоко", "b": "молоток", "expected": 2},
    {"a": "стол", "b": "стул", "expected": 1},
    {"a": "кот", "b": "кит", "expected": 1},
    {"a": "работа", "b": "работник", "expected": 3},
    {"a": "осень", "b": "весна", "expected": 4},
    {"a": "собака", "b": "кошка", "expected": 3},
    {"a": "мир", "b": "мирный", "expected": 3},
]


def test_levenshtein_mat_rs() -> None:
    for case in cases:
        assert levenshtein_mat_rs(case["a"], case["b"]) == case["expected"]


def test_levenshtein_vec_rs() -> None:
    for case in cases:
        assert levenshtein_vec_rs(case["a"], case["b"]) == case["expected"]


def test_levensthein_tweaked_rs() -> None:
    for case in cases:
        assert levenshtein_tweaked_rs(case["a"], case["b"]) == case["expected"]


def test_levenshtein_jelly() -> None:
    for case in cases:
        assert levenshtein_jelly(case["a"], case["b"]) == case["expected"]


def test_levenshtein_mat_py() -> None:
    for case in cases:
        assert levenshtein_mat_py(case["a"], case["b"]) == case["expected"]


def test_levenshtein_vec_py() -> None:
    for case in cases:
        assert levenshtein_vec_py(case["a"], case["b"]) == case["expected"]


def test_levenshtein_exp_py() -> None:
    for case in cases:
        assert levenshtein_exp_py(case["a"], case["b"]) == case["expected"]


def test_levenshtein_cy() -> None:
    for case in cases:
        assert levenshtein_cy(case["a"], case["b"]) == case["expected"]



from fxkk import (
    levenshtein_mat_py,
    levenshtein_vec_py,
    levenshtein_exp_py,
    jellyfish_levenshtein as jellyfish_raw,
    levenshtein_cy,
    levenshtein_mat_rs,
    levenshtein_vec_rs,
    levenshtein_tweaked_rs,
)
from jellyfish import levenshtein_distance as jellyfish_compiled
from typing import Any
import random
import string


def random_string(n: int = 10) -> str:
    return "".join(random.choice(string.ascii_letters) for _ in range(n))


cases = [(random_string(), random_string()) for _ in range(1000)]


def test_levenshtein_mat_rs_performance(benchmark: Any) -> None:
    def profile_levenshtein() -> None:
        for a, b in cases:
            levenshtein_mat_rs(a, b)

    benchmark(profile_levenshtein)


def test_levenshtein_vec_rs_performance(benchmark: Any) -> None:
    def profile_levenshtein() -> None:
        for a, b in cases:
            levenshtein_vec_rs(a, b)

    benchmark(profile_levenshtein)


def test_levenshtein_tweaked_rs_performance(benchmark: Any) -> None:
    def profile_levenshtein() -> None:
        for a, b in cases:
            levenshtein_tweaked_rs(a, b)

    benchmark(profile_levenshtein)


def test_jellyfish_raw_performance(benchmark: Any) -> None:
    def profile_levenshtein() -> None:
        for a, b in cases:
            jellyfish_raw(a, b)

    benchmark(profile_levenshtein)


def test_jellyfish_compiled_performance(benchmark: Any) -> None:
    def profile_levenshtein() -> None:
        for a, b in cases:
            jellyfish_compiled(a, b)

    benchmark(profile_levenshtein)


def test_levenshtein_mat_performance(benchmark: Any) -> None:
    def profile_fast_levenshtein() -> None:
        for a, b in cases:
            levenshtein_mat_py(a, b)

    benchmark(profile_fast_levenshtein)


def test_levenshtein_vec_performance(benchmark: Any) -> None:
    def profile_levenshtein_vec() -> None:
        for a, b in cases:
            levenshtein_vec_py(a, b)

    benchmark(profile_levenshtein_vec)


def test_levenshtein_cy_performance(benchmark: Any) -> None:
    def profile_levenshtein() -> None:
        for a, b in cases:
            levenshtein_cy(a, b)

    benchmark(profile_levenshtein)


def test_levenstein_exp_performance(benchmark: Any) -> None:
    def profile_levenshtein() -> None:
        for a, b in cases:
            levenshtein_exp_py(a, b)

    benchmark(profile_levenshtein)



from .levenshtein_cy import levenshtein as levenshtein_cy
from ._levenshtein import (
    levenshtein_mat_py,
    levenshtein_exp_py,
    levenshtein_vec_py,
    jellyfish_levenshtein,
)
from fxkk.fxkk import (
    levenshtein_mat as levenshtein_mat_rs,
    levenshtein_vec as levenshtein_vec_rs,
    levenshtein_tweaked as levenshtein_tweaked_rs,
)



from fxkk import (
    levenshtein_mat_py,
    levenshtein_vec_py,
    levenshtein_exp_py,
    jellyfish_levenshtein as jellyfish_raw,
    levenshtein_cy,
    levenshtein_mat_rs,
    levenshtein_vec_rs,
    levenshtein_tweaked_rs,
)
from jellyfish import levenshtein_distance as jellyfish_compiled
from typing import Any
import random
import string


def random_string_shared_fixes(n: int = 10) -> str:
    return (
        "gillespie "
        + "".join(random.choice(string.ascii_letters) for _ in range(n))
        + " theorem"
    )


cases_shared = [
    (random_string_shared_fixes(), random_string_shared_fixes()) for _ in range(1000)
]


def test_levenshtein_tweaked_rs_performance_shared_fixes(benchmark: Any) -> None:
    def profile_levenshtein() -> None:
        for a, b in cases_shared:
            levenshtein_tweaked_rs(a, b)

    benchmark(profile_levenshtein)


def test_jellyfish_performance_shared_fixes(benchmark: Any) -> None:
    def profile_levenshtein() -> None:
        for a, b in cases_shared:
            jellyfish_compiled(a, b)

    benchmark(profile_levenshtein)



# type: ignore
from typing import List
import array as arrayModule


def levenshtein_mat_py(a: str, b: str) -> int:
    """Calculate the Levenshtein distance between two strings."""
    if a == b:
        return 0
    if not a:
        return len(b)
    if not b:
        return len(a)
    l_a = len(a) + 1
    l_b = len(b) + 1
    if l_a == 0:
        return l_b
    if l_b == 0:
        return l_a
    if a == b:
        return 0

    D: List[List[int]] = []
    for i in range(l_a):
        D.append([0] * l_b)

    for i in range(1, l_a):
        D[i][0] = i

    for j in range(1, l_b):
        D[0][j] = j

    for i in range(1, l_a):
        a_i = a[i - 1]
        for j in range(1, l_b):
            b_j = b[j - 1]
            cost = 0 if a_i == b_j else 1
            D[i][j] = min(
                D[i - 1][j] + 1,  # deletion
                D[i][j - 1] + 1,  # insertion
                D[i - 1][j - 1] + cost,  # substitution
            )
    return D[l_a - 1][l_b - 1]


def jellyfish_levenshtein(s1, s2):
    if s1 == s2:
        return 0
    rows = len(s1) + 1
    cols = len(s2) + 1

    if not s1:
        return cols - 1
    if not s2:
        return rows - 1

    prev = None
    cur = range(cols)
    for r in range(1, rows):
        prev, cur = cur, [r] + [0] * (cols - 1)
        for c in range(1, cols):
            deletion = prev[c] + 1
            insertion = cur[c - 1] + 1
            edit = prev[c - 1] + (0 if s1[r - 1] == s2[c - 1] else 1)
            cur[c] = min(edit, deletion, insertion)

    return cur[-1]


def levenshtein_vec_py(s: str, t: str) -> int:
    if s == t:
        return 0
    if not s:
        return len(t)
    if not t:
        return len(s)

    l_s = len(s) + 1
    l_t = len(t) + 1
    v_0 = None
    v_1 = range(l_t)

    for i in range(1, l_s):
        # Old becomes new, new gets initialized to deletion_length
        v_0, v_1 = v_1, [i] + [0] * (l_t - 1)
        for j in range(1, l_t):
            # Calculate the cost of deleting, inserting, or substituting
            deletion = v_0[j] + 1
            insertion = v_1[j - 1] + 1
            substitution = v_0[j - 1] + (0 if s[i - 1] == t[j - 1] else 1)
            v_1[j] = min(deletion, insertion, substitution)
    return v_1[-1]


def levenshtein_exp_py(s, t):
    # https://turnerj.com/blog/levenshtein-distance-part-2-gotta-go-fast
    if s == t:
        return 0
    if not s:
        return len(t)
    if not t:
        return len(s)
    previous_row = arrayModule.array("I", [1] * (len(t) + 1))
    for i in range(1, len(s) + 1):
        previous_diagonal = i - 1
        previous_column = i
        sourceChar = s[i - 1]
        for j in range(1, len(t) + 1):
            local_cost = previous_diagonal
            deletion_cost = previous_row[j]
            if sourceChar != t[j - 1]:
                local_cost = min(local_cost, deletion_cost, previous_column) + 1
            previous_column = local_cost
            previous_row[j] = local_cost
            previous_diagonal = deletion_cost
    return previous_row[-1]


def levenshtein_exp_with_arr(s, t):
    # https://turnerj.com/blog/levenshtein-distance-part-2-gotta-go-fast
    if s == t:
        return 0
    if not s:
        return len(t)
    if not t:
        return len(s)
    previous_row = arrayModule.array("I", range(len(t) + 1))
    for i in range(1, len(s) + 1):
        previous_diagonal = i - 1
        previous_column = previous_diagonal + 1
        sourceChar = s[i - 1]
        for j in range(1, len(t) + 1):
            local_cost = previous_diagonal
            deletion_cost = previous_row[j]
            if sourceChar != t[j - 1]:
                local_cost = min(local_cost, deletion_cost, previous_column) + 1
            previous_column = local_cost
            previous_row[j] = local_cost
            previous_diagonal = deletion_cost
    return previous_row[-1]


def levenshtein_turner(s, t):
    # https://turnerj.com/blog/levenshtein-distance-part-2-gotta-go-fast
    if s == t:
        return 0
    if not s:
        return len(t)
    if not t:
        return len(s)
    previous_row = list(range(len(t) + 1))
    for i in range(1, len(s) + 1):
        previous_diagonal = i - 1
        previous_column = previous_diagonal + 1
        sourceChar = s[i - 1]
        for j in range(1, len(t) + 1):
            local_cost = previous_diagonal
            deletion_cost = previous_row[j]
            if sourceChar != t[j - 1]:
                local_cost = min(local_cost, deletion_cost, previous_column) + 1
            previous_column = local_cost
            previous_row[j] = local_cost
            previous_diagonal = deletion_cost
    return previous_row[-1]


def levenshtein_exp_backup(s, t):
    # https://turnerj.com/blog/levenshtein-distance-part-2-gotta-go-fast
    if s == t:
        return 0
    if not s:
        return len(t)
    if not t:
        return len(s)
    previous_row = list(range(len(t) + 1))
    for i in range(1, len(s) + 1):
        previous_diagonal = i - 1
        previous_column = previous_diagonal + 1
        sourceChar = s[i - 1]
        for j in range(1, len(t) + 1):
            if sourceChar == t[j - 1]:
                previous_column = previous_diagonal
            else:
                previous_column = (
                    min(previous_column, previous_diagonal, previous_row[j]) + 1
                )
            previous_diagonal = previous_row[j]
            previous_row[j] = previous_column
    return previous_row[-1]



from setuptools import setup
from Cython.Build import cythonize

setup(
    ext_modules=cythonize("./fxkk/*.pyx", annotate=True, build_dir="build"),
    options={"build_ext": {"build_lib": "./fxkk"}},
)
